#!/usr/bin/perl -w

## Finding the TreeTagger binaries and file: this script tries to
## find all necessary TreeTagger files using environment variables,
## parameters, or the located of the treetagger program if found in the
## current PATH (see run-treetagger.pl -h)
## Only if all this fails, you can set the location 
## in here by replacing the undef (no quotes) with the full pathname
## to the corresponding directory (in single or double quotes).
our $HOME_TreeTagger = undef; 
our $BIN_TreeTagger = undef;
our $CMD_TreeTagger = undef;
our $LIB_TreeTagger = undef;
###############################################################
### NO NEED TO CHANGE ANYTHING BELOW THIS LINE!!            ###
###############################################################

use strict;
my $version = '';

use Getopt::Long;
use File::Spec;
use FindBin;
use File::Basename;
use Cwd 'abs_path';
use Pod::Usage;

our $PRV = "INFO(run-treetagger.pl):";
our $PRD = "DEBUG(run-treetagger.pl):";

my %options = ();

GetOptions(\%options, 
           "bin=s", "cmd=s","lib=s","home=s","chunk","lemma","tmpdir=s",
           "lang=s",
           "v","d","h","man") or pod2usage(2);
if(defined($options{h})) {
  pod2usage(0);
  exit(0);
}
if(defined($options{man})) {
  pod2usage(-verbose => 2);
  exit(0);
}

our $debug = defined($options{d});
our $verbose = defined($options{v}) || $debug; ## debug implies verbose

our $lemmas4chunking = $options{lemma} || "";

if(!$BIN_TreeTagger) {
  $BIN_TreeTagger ||= $options{"bin"};
}
if(!$BIN_TreeTagger) {
  $HOME_TreeTagger ||= $options{"home"} unless($HOME_TreeTagger);
  $BIN_TreeTagger = File::Spec->catfile($HOME_TreeTagger,"bin") if($HOME_TreeTagger);
}
if(!$BIN_TreeTagger) {
  $BIN_TreeTagger = $ENV{"TREETAGGER_BIN"};
}
if(!$BIN_TreeTagger) {
  $HOME_TreeTagger = $ENV{"TREETAGGER_HOME"} unless($HOME_TreeTagger);
  $BIN_TreeTagger = File::Spec->catfile($HOME_TreeTagger,"bin") if($HOME_TreeTagger);
}
if(!$BIN_TreeTagger) {
  $BIN_TreeTagger = `which treetagger`;
  $BIN_TreeTagger = File::Basename::dirname($BIN_TreeTagger) if($BIN_TreeTagger);
  $HOME_TreeTagger = Cwd::abs_path($BIN_TreeTagger . "..") if($BIN_TreeTagger);
}
die "Could not determine location of TreeTagger bin directory" unless($BIN_TreeTagger);
print STDERR "$PRV using bin directory $BIN_TreeTagger\n" if($verbose);

if(!$CMD_TreeTagger) {
  $CMD_TreeTagger ||= $options{"cmd"};
}
if(!$CMD_TreeTagger) {
  $HOME_TreeTagger ||= $options{"home"} unless($HOME_TreeTagger);
  $CMD_TreeTagger = File::Spec->catfile($HOME_TreeTagger,"cmd") if($HOME_TreeTagger);
}
if(!$CMD_TreeTagger) {
  $CMD_TreeTagger = $ENV{"TREETAGGER_CMD"};
}
if(!$CMD_TreeTagger) {
  $HOME_TreeTagger = $ENV{"TREETAGGER_HOME"} unless($HOME_TreeTagger);
  $CMD_TreeTagger = File::Spec->catfile($HOME_TreeTagger,"cmd") if($HOME_TreeTagger);
}
if(!$CMD_TreeTagger) {
  my $tmp = `which treetagger`;
  $tmp = File::Basename::dirname($tmp) if($tmp);
  $HOME_TreeTagger = Cwd::abs_path($tmp . "..") if($tmp);
  $CMD_TreeTagger = File::Spec->catfile($HOME_TreeTagger,"cmd") if($HOME_TreeTagger);
}
die "Could not determine location of TreeTagger cmd directory" unless($CMD_TreeTagger);
print STDERR "$PRV using cmd directory $CMD_TreeTagger\n" if($verbose);

if(!$LIB_TreeTagger) {
  $LIB_TreeTagger ||= $options{"lib"};
}
if(!$LIB_TreeTagger) {
  $HOME_TreeTagger ||= $options{"home"} unless($HOME_TreeTagger);
  $LIB_TreeTagger = File::Spec->catfile($HOME_TreeTagger,"lib") if($HOME_TreeTagger);
}
if(!$LIB_TreeTagger) {
  $LIB_TreeTagger = $ENV{"TREETAGGER_LIB"};
}
if(!$LIB_TreeTagger) {
  $HOME_TreeTagger = $ENV{"TREETAGGER_HOME"} unless($HOME_TreeTagger);
  $LIB_TreeTagger = File::Spec->catfile($HOME_TreeTagger,"lib") if($HOME_TreeTagger);
}
if(!$LIB_TreeTagger) {
  my $tmp = `which treetagger`;
  $tmp = File::Basename::dirname($tmp) if($tmp);
  $HOME_TreeTagger = Cwd::abs_path($tmp . "..") if($tmp);
  $LIB_TreeTagger = File::Spec->catfile($HOME_TreeTagger,"lib") if($HOME_TreeTagger);
}
die "Could not determine location of TreeTagger lib directory" unless($LIB_TreeTagger);
print STDERR "$PRV using lib directory $LIB_TreeTagger\n" if($verbose);

my $tmpdir = undef;
if($options{"tmpdir"}) {
  $tmpdir = $options{"tmpdir"};
}
unless($tmpdir) {
  $tmpdir = $ENV{"TMP"};
}
if($lemmas4chunking && !$tmpdir) {
  die "Need a temporary directory for -lemma but neither -tmp option nor \$TMP environment variable present";
}

our $lang = $options{"lang"} || "english";
our $tagger_bin = File::Spec->catfile($BIN_TreeTagger,"tree-tagger");

if($options{"chunk"}) {
  print STDERR "$PRD doing chunking ...\n" if($debug);
  my $parfile1 = File::Spec->catfile($LIB_TreeTagger,$lang.".par");
  my $infiles = join(" ",@ARGV);
  my $posparms = "-quiet -token -lemma -sgml -pt-with-lemma -ignore-prefix ";
  my $cmd = "cat $infiles | $tagger_bin $parfile1 $posparms ";
  my $tmpfile1 = "";
  if($lemmas4chunking) {
    $tmpfile1 = File::Spec->catfile($tmpdir,"run-treetagger_tmp1_$$.txt");
    $cmd .= " | tee $tmpfile1 ";
  }
  $cmd .= ' | perl -nae \'if ($#F==0){print}else{print "$F[0]-$F[1]\n"}\' ';
  my $parfile2 = File::Spec->catfile($LIB_TreeTagger,$lang."-chunker.par");
  my $chunkparms = " -token -sgml -hyphen-heuristics -quiet -eps 0.00000001";
  $cmd .= " | $tagger_bin $parfile2 $chunkparms";
  my $tmpfile2 = "";
  if($lemmas4chunking) {
    $tmpfile2 = File::Spec->catfile($tmpdir,"run-treetagger_tmp2_$$.txt");
    $cmd .= " > $tmpfile2 ";
  }
  print STDERR "$PRD Running: $cmd\n" if($debug);
  system($cmd);
  if($lemmas4chunking) {
    $cmd = "perl -nae 'print \$F[2],\"\n\";' $tmpfile1 ";
    $cmd .= " | paste $tmpfile2 -";
    print STDERR "$PRD Running: $cmd\n" if($debug);
    system($cmd);
  unless($debug) {
    unlink($tmpfile1) if($tmpfile1);
    unlink($tmpfile2) if($tmpfile2);
  }
  }
} else {
  print STDERR "$PRD doing POS tagging ...\n" if($debug);
  ## default is to do POS tagging
  my $parfile = File::Spec->catfile($LIB_TreeTagger,$lang.".par");
  my $infiles = join(" ",@ARGV);
  my $posparms = "-quiet -token -lemma -sgml";
  my $cmd = "cat $infiles | $tagger_bin $parfile $posparms";
  if($lang eq "german") {
    my $filter = File::Spec->catfile($CMD_TreeTagger,"filter-german-tags");
    $cmd .= " | $filter";
  }
  print STDERR "$PRD Running: $cmd\n" if($debug);
  system($cmd);
}

__END__

=head1 NAME

run-treetagger.pl - nice interface to call the TreeTagger 

=head1 DESCRIPTION

This program provides a comfortable interface to the TreeTagger, allowing
several ways to find the TreeTagger binaries and parameter files so that
this file does not need to get modified (not good to modify something in
a standard GATE plugin directory).

The program will use the following ways to set the location of the bin, cmd
and lib subdirectories of the TreeTagger in decreasing order or priority:
the value of the variables defined at the top of this script,
the command line options -bin/-cmd/-lib, the environment variable
TREETAGGER_BIN, TREETAGGER_CMD, and TREETAGGER_LIB, the environment variable
TREETAGGER_HOME (used as the common parent of these three directories), and
the parent directory of the location of where in the path the binary file 
treetagger is found as the common parent of these three directories.

The script can be used to initate both POS tagging and chunking and will
look for a par file that corresponds to the language and task in the lib
directory according to the following naming scheme: for POS tagging, look 
for a file <language>.par. For chunking, look for a file <language>-chunk.par.

The script will automatically do the necessary postprocessing for German.

(NOTE: this script probably needs to get extended to handle languages
other than English and German correctly or better).

=head1 SYNOPSIS

  perl run-treetagger.pl -h | -man

  perl run-treetagger.pl [-d] [-v] 
    [-bin bindir] [-cmd cmddir] [-lib libdir] [-home homedir]
    [-chunk] [-lang {english|german|french|..}]
    [-lemma] [-tmpdir]
    [files ...]

=head2 OPTIONS

=over 8

=item B<-h>

Show usage information.

=item B<-man>

Show the complete manual page.

=item B<-d>

Switch on debugging messages.

=item B<-v>

Switch on verbose messages.

=item B<-lang {english|german|french|...}>

The language-specific model to use. The language-specific model files are
located in the TreeTagger "lib" subdirectory and start with the language
name followed by a hyphen, followed by e.g. ".par" or "-chunker.par". 
Specify the language name (without the hyphen) here.

=item B<-bin bindir>

Set the directory where to find the TreeTagger binary.

=item B<-cmd cmddir>

Set the directory where to find the TreeTagger shell scripts
(NOTE: no shell scripts are used in this script yet so this directory 
does not need to be set yet).
 
=item B<-lib libdir>

Set the location of the TreeTagger lib directory which contains all
the parameter (.par) files.

=item B<-home homedir>

Set the root directory of the TreeTagger installation. This directory
should contain the "bin", "cmd", and "lib" subdirectories..

=item B<-chunk>

Run chunking. If not specified, run POS tagging. Chunking is done by 
first running the POS tagger using the <language>.par parameter file,
then running the treetagger again on columns 1 and 2 of the POS tagged output,
using the <language>-chunk.par parameter file.
The lemma of the POS phase can optionally be appended to this output with
the -lemma option.
The chunker output is not postporcessed to produce sgml-type annotations. 
The original chunk tags per token as provided by the treetagger are
returned.

=item B<-lemma>

Only relevant if -chunk is also supplied. Will add a column containing the
lemma from the POS phase to the chunker output. 

=item B<-tmpdir dirname>

The directory that should be used for storing temporary files. Temporary
files will only be created if the option -lemma is used.
NOTE: the temporary files (two per run) will NOT be deleted by this script!
The files will be named run-treetagger_tmp<N>_<P>.txt where <N> is a number
(currently only 1 and 2) and <P> is the numeric process id.

The caller of this script should take care of removing these temporary files.

=back

=cut
