package gate.treetagger2;

import java.io.*;
import java.nio.charset.*;
import java.util.*;
import java.net.URL;

import gate.*;
import gate.creole.*;
import gate.util.*;

/**
 * This class is a wrapper for the language-independent POS tagger from the
 * University of Stuttgart, Germany.
 * It passes GATE a document and a Tree Tagger shell script.
 * Results are stored in the document's TreeTaggerToken annotations
 */

public abstract class TreeTaggerBase
    extends AbstractLanguageAnalyser
    implements ProcessingResource {

  static final long serialVersionUID = 20060730122746001L;
  
  /*  The location of where the script for invoking the TreeTagger
   *  is stored. The location is determined automatically and checked
   *  at initialization time.
   */
  private File runTreeTaggerScript;
  
  /**
   * Initialize this resource.  
   * Find the location of the script for invoking the TreeTagger.
   * Check if there is a temporary directory.
   */
  public Resource init() throws ResourceInstantiationException {
    
    CreoleRegister reg = Gate.getCreoleRegister();
    ResourceData myPR = (ResourceData) reg.get("gate.treetagger2.TreeTaggerPOS");
    if(myPR == null) {
      throw new ResourceInstantiationException(
          "Could not find the ResourceData for gate.treetagger2.TreeTaggerPOS"
          );
    }
    URL myPRURL = myPR.getXmlFileUrl();
    java.io.File myPRFile = gate.util.Files.fileFromURL(myPRURL);
    File cmdDir = new File(myPRFile.getParentFile(),"cmd");
    runTreeTaggerScript = new File(cmdDir,"run-treetagger.pl"); 
    
    String tmpDir = System.getProperty("java.io.tmpdir");
    if (tmpDir == null || tmpDir.indexOf(' ') >= 0) {
      throw new ResourceInstantiationException(
          "TreeTagger requires your temporary directory to be set to a value "
          + "that does not contain spaces.  Please set java.io.tmpdir to a "
          + "suitable value.");
    }    
    return this;
  }

  /**
   * Run the TreeTagger on the current document.  This writes the document text
   * to a temporary file, runs the tagger and processes its output to produce
   * TreeTaggerToken annotations on the document.
   */
  public void execute() throws ExecutionException {
    // needs: document, charset, runTreeTaggerScript
    if (document == null)
      throw new ExecutionException("No document to process!");

    if (charset == null) {
      throw new ExecutionException("No encoding specified");
    }

    //get current text from GATE for the TreeTagger
    File textfile = getCurrentText();
    
    // check that the file exists
    File scriptfile = runTreeTaggerScript;
    if (scriptfile.exists()==false)
      throw new ExecutionException("Script "+scriptfile.getAbsolutePath()+" does not exist");
    
    //generate TreeTagger command line
    String debugOption = "";
    if(getDebugMode().booleanValue()) {
      debugOption = " -d ";
    }
    String treeTaggerCmd = 
        "perl " +
        scriptfile.getAbsolutePath() + " " +
        debugOption +
        treeTaggerInvocationScriptParms + " " +
        addToScriptParms + " " +
        textfile.getAbsolutePath();
    
    if(getDebugMode().booleanValue()) {
      // in debug mode, copy the output from the script to a temporary file
      File debugFile;
      try {
        debugFile = File.createTempFile("treetagger_out", ".txt");
      } catch(java.io.IOException e) {
        throw (ExecutionException)new ExecutionException(
        "Error creating temporary file for TreeTagger")
        .initCause(e);
      }
      treeTaggerCmd = treeTaggerCmd + " | tee " +
        debugFile.getAbsolutePath();
    }
    if(debugMode) {
      System.out.println("Running: "+treeTaggerCmd);
    }
    //run TreeTagger
    runTreeTagger(treeTaggerCmd);

    //delete the temporary text file, unless we want to keep it
    if(!getDebugMode().booleanValue()) {
      textfile.delete();
    }
  }

  private File getCurrentText() throws ExecutionException {
    File gateTextFile = null;
    try {
      gateTextFile = File.createTempFile("treetagger_in", ".txt");
      // depending on the failOnUnmappableChar parameter, we either make the
      // output stream writer fail or replace the unmappable character with '?'
      CharsetEncoder charsetEncoder = charset.newEncoder()
          .onUnmappableCharacter(
              failOnUnmappableChar ? CodingErrorAction.REPORT
                                   : CodingErrorAction.REPLACE);
      FileOutputStream fos = new FileOutputStream(gateTextFile);
      OutputStreamWriter osw = new OutputStreamWriter(fos, charsetEncoder);
      BufferedWriter bw = new BufferedWriter(osw);
      AnnotationSet annotSet = (annotationSetName == null || annotationSetName.trim().length() == 0) ?
          document.getAnnotations() : document.getAnnotations(annotationSetName);
      // was: annotSet = annotSet.get(ANNIEConstants.TOKEN_ANNOTATION_TYPE);
      annotSet = annotSet.get(tokenAnnotationType);
      if(annotSet == null || annotSet.size() == 0) {
        throw new GateRuntimeException("No Gate Tokens found in the document.. please run tokenizer first or specify the correct type of the token annotations");
      }

      // sort tokens according to their offsets
      ArrayList tokens = new ArrayList(annotSet);
      Collections.sort(tokens, new OffsetComparator());
      // and now start writing them in a file
      for(int i=0;i<tokens.size();i++) {
        FeatureMap features = ((Annotation) tokens.get(i)).getFeatures();
        if(features == null || features.size() == 0 || !features.containsKey(gate.creole.ANNIEConstants.TOKEN_STRING_FEATURE_NAME)) {
          throw new GateRuntimeException("GATE Tokens must have string feature, which couldn't be found");
        }
        String string = (String) features.get(gate.creole.ANNIEConstants.TOKEN_STRING_FEATURE_NAME);
        bw.write(string);
        if(i+1 < tokens.size())
          bw.newLine();
      }
      bw.close();
      
    }
    catch (CharacterCodingException cce) {
      throw (ExecutionException)new ExecutionException(
          "Document contains a character that cannot be represented " +
          "in " + encoding)
          .initCause(cce);
    }
    catch (java.io.IOException except) {
      throw (ExecutionException)new ExecutionException(
          "Error creating temporary file for TreeTagger")
          .initCause(except);
    }
    return (gateTextFile);
  }

  private void runTreeTagger(String cmdline) throws ExecutionException {
    String line;
    AnnotationSet aSet;

    aSet = (annotationSetName == null || annotationSetName.trim().length() == 0) ?
        document.getAnnotations() : document.getAnnotations(annotationSetName);

    CharsetDecoder charsetDecoder = charset.newDecoder();
    // run TreeTagger and save output
    try {

      Process p = Runtime.getRuntime().exec(cmdline);

      //get its tree tagger output (gate input)
      BufferedReader input =
          new BufferedReader
          (new InputStreamReader(p.getInputStream(), charsetDecoder));

      // let us store all tagged data in lines
      // there must be at leat the original form a tab a POS
      // and possibly a lemma preceded by a tab
      ArrayList lines = new ArrayList();
      while( (line = input.readLine()) != null) {
        if(line.split("\t").length>1)
          lines.add(line);
      }

      // we need to check if the tokenization was done correctly
      // and we do so it by comparing the number of lines with the number of
      // tokens created by the GATE tokenizer
      //AnnotationSet annotSet = aSet.get(ANNIEConstants.TOKEN_ANNOTATION_TYPE);
      AnnotationSet annotSet = aSet.get(tokenAnnotationType);
      if(annotSet == null || annotSet.size() == 0) {
        throw new GateRuntimeException("No Gate Tokens found in the document.. please run tokenizer first or specify correct Token annotation type");
      }

      // sort tokens according to their offsets
      ArrayList tokens = new ArrayList(annotSet);
      Collections.sort(tokens, new OffsetComparator());


      if(tokens.size() != lines.size()) {
        System.out.println("Tokens : "+tokens.size()+" lines : "+lines.size());
        throw new GateRuntimeException("Document does not have the expected number of tokens created by the treeTaggerBinary file");
      }

      // Create the features from the lines read in
      getFeatures4Tokens(lines, tokens);
    }
    catch (Exception err) {
      throw (ExecutionException)new ExecutionException(
          "Error occurred running TreeTagger")
          .initCause(err);
    }

  }

  abstract protected void getFeatures4Tokens (ArrayList lines, ArrayList tokens);
  
  
  // getter and setter methods and their associated fields

  /**
   * Set the document to process.
   */
  public void setDocument(gate.Document document) {
    this.document = document;
  }

  /**
   * Return the document being processed.
   */
  public gate.Document getDocument() {
    return document;
  }

  private gate.Document document;

  /**
   * Set the name of the annotation set to place the generated TreeTaggerToken
   * annotations in.
   */
  public void setAnnotationSetName(String annotationSetName) {
    this.annotationSetName = annotationSetName;
  }

  /**
   * Return the annotation set name used for the TreeTaggerTokens.
   */
  public String getAnnotationSetName() {
    return annotationSetName;
  }

  protected String annotationSetName;

  /**
   * Set the location of the TreeTagger script.
   */
  public void setTreeTaggerInvocationScriptParms(String treeTaggerInvocationScriptParms) {
    this.treeTaggerInvocationScriptParms = treeTaggerInvocationScriptParms;
  }

  /**
   * Return the location of the TreeTagger script.
   */
  public String getTreeTaggerInvocationScriptParms() {
    return treeTaggerInvocationScriptParms;
  }

  protected String treeTaggerInvocationScriptParms = "";

  protected String addToScriptParms = "";
  /**
   * Set the character encoding to use for the temporary files.  This must be
   * the encoding that your tree tagger understands.
   *
   * @throws IllegalCharsetNameException if the specified string is not a valid
   * encoding name.
   */
  public void setEncoding(String newEncoding) {
    this.charset = Charset.forName(newEncoding);
    // if the line above didn't throw an exception, we are OK.
    this.encoding = newEncoding;
  }

  /**
   * Get the character encoding used for the temporary files.
   */
  public String getEncoding() {
    return encoding;
  }
  private Charset charset;
  protected String encoding;

  /**
   * Set the flag for whether we should fail if an unmappable character is
   * found.
   */
  public void setFailOnUnmappableChar(Boolean newValue) {
    failOnUnmappableChar = (newValue == null) ? true : newValue.booleanValue();
  }

  /**
   * Get the flag for whether we should fail if an unmappable character is
   * found.
   */
  public Boolean getFailOnUnmappableChar() {
    return Boolean.valueOf(failOnUnmappableChar);
  }
  
  protected boolean failOnUnmappableChar;

  public void setDebugMode(Boolean value) {
    debugMode = (value == null) ? false : value.booleanValue();
  }
  public Boolean getDebugMode() {
    return Boolean.valueOf(debugMode);
  }
  
  public String getTokenAnnotationType() {
    return tokenAnnotationType;
  }
  
  protected String tokenAnnotationType;
  
  public void setTokenAnnotationType(String type) {
    tokenAnnotationType = type;
  }
  
  protected boolean debugMode;
} // class TreeTaggerPOS
