package gate.treetagger2;

import java.util.*;

import gate.*;
import gate.creole.*;
import gate.util.*;


/** This class is a wrapper for the language-independent POS tagger from the
 * University of Stuttgart, Germany.
 * 
 * This class is for the chunker function of TreeTagger.
 * 
 * This is a modified version of the plugin that comes with GATE version
 * 3.1 and earlier. This modified version includes several changes 
 * and enhancements over the original version:
 * <ul>
 * <li>The user does not have to specify the full path to the script
 * that runs TreeTagger. Instead, the plugin derives the location from
 * the location where the plugin is stored.
 * <li>The script that runs TreeTagger nearly never ever needs to be
 * changed since it uses several ways of determining how to invocate
 * the TreeTagger binaries and where to locate the TreeTagger parameter 
 * files (see below).
 * <li>The plugin does not read back the temporary TreeTagger input file
 * and store its content in temporary features of the Token annotations as 
 * the old version did, thus increasing CPU and memory efficiency a bit. 
 * The only check for consistency that is now done is
 * by comparing the number of Tokens with the number of lines read back
 * from the TreeTagger output which should be sufficient. 
 * <li>All plugin parameters are now runtime parameters so they can be
 * changed after creating the processing resource much more easily.
 * <li>The plugin now provides two processing resources: one for POS tagging
 * and one for chunking (provided a chunking model is available). This 
 * class handles the chunker.
 * <li>The plugin allows to specify an annotation type other than Token to
 * be used as input tot the TreeTagger program.
 * </ul>
 * 
 * The following parameters are available for the TreeTaggerPOS part of
 * speech tagger processing resource:
 * <ul>
 * <li> annotationSetName: the name of the annotation set to process</li>
 * <li> debugMode: true/false -- if true, will keep temporary files and show
 *   debugging messages</li>
 * <li> doPosTagging: if true (the default), this invocation of the TreeTagger
 * will generate both POS tags and the chunking information.</li> 
 * <li> encoding: the encoding to use when creating the file that will be
 *   read in by the TreeTagger binary. This defaults to ISO-8859-1 and this
 *   is currently the only encoding supported by TreeTagger. </li>
 * <li> failOnUnmappableChar: what to do if the document contains a 
 *   character that cannot be converted to the specified encoding. If true,
 *   the processing will be interrupted, if false, the character will be 
 *   silently ignored.</li>
 * <li>includeLemma: if true (default is false), lemmata will be included in
 *   the POS features.</li>
 * <li> tokenAnnotationType: the type name of the tokens to be used for 
 *   POS tagging. The default is "Token".</li>
 * <li> treeTaggerInvocationScriptParms: everything specified here is 
 *   passed on to the invocation script that actually locates and invokes
 *   the TreeTagger binary. The invocation script is located in the 
 *   plugin directory, subdirectory <code>cmd</code> and has the 
 *   name <code>run-treetagger.pl</code>.
 *   It can be invoked manually with the command 
 *   <code>perl cmd/run-treetagger.pl -h</code> to show all possible command
 *   line options and with the command 
 *   <code>perl cmd/run-treetagger.pl -man</code> to show the full documentation. 
 *   Note that some arguments do not apply to POS tagging and that 
 *   some arguments are passed automatically: <code>-tmpdir</code> and 
 *   <code>-chunk</code>. The only 
 *   argument that needs to be specified in most cases is <code>-lang</code>.
 *   </li>
 * </ul>
 * <p>
 * The output of the chunker is stored in tree additional features for each
 * annotation: 
 * <code>chunktag</code> which holds the original tag assigned from the
 * TreeTagger, <code>chunkpart</code> which contains the sequence 
 * information from the chunktag and <code>chunktype</code> which contains
 * the type information from the chunktag.
 * <p>
 * In order to create annotations that span several tokens of the same type,
 * you need postprocess the output of the chunker with a JAPE transducer.
 * An example JAPE rule file that will work with the default token annotation
 * type is provided in the plugin directory, subdirectory 
 * <code>resources/grammar</code>
 * as file <code>join.jape</code>.
 *
 *  @author Johann Petrak, Austrian Research Institute for AI (OFAI)
 */

public class TreeTaggerChunk
    extends TreeTaggerBase
    implements ProcessingResource {

  static final long serialVersionUID = 20060730122746003L;
  
  public void execute() throws ExecutionException {
    // use addToScriptParms
    if(getIncludeLemma().booleanValue()) {
      String tmpDir = System.getProperty("java.io.tmpdir");
      addToScriptParms =  "-lemma -chunk -tmp " + tmpDir;
    } else {
      addToScriptParms = "-chunk"; 
    }
    super.execute();
  }
  private String treeTaggerInvocationScriptParms;
  
  /** Parse the lines of TreeTagger Chunker output and create features for
   *  the tokens. 
   *  The chunker output has the format:
   *     word-POS   POS/CHUNK
   *  If the doPosTagging flag is true, we also set the POS tags from
   *  this output (using the first part of the POS/CHUNK pair)
   *  If the includeLemma flag is true, the invocation script is called 
   *  with the -lemma option. In this case, we will set 
   *  tje lemma in addition to the chunk tags. The format in this case is:
   *     word-POS  POS/CHUNK LEMMA  
   */
  protected void getFeatures4Tokens (ArrayList lines, ArrayList tokens) {
    String line, word1, word2, pos, chunk, part, type;
    String[] word2parts;
    for(int i=0;i<lines.size();i++) {
      line = (String) lines.get(i);
      String[] words = line.split("\\s+");
      if((getIncludeLemma().booleanValue() && words.length != 3) ||
         (!getIncludeLemma().booleanValue() && words.length != 2)) {
        throw new GateRuntimeException("A line returned by the TreeTagger does not contain the correct number of tokens: \""+line+"\"");
      }
      word1 = words[0];
      word2 = words[1];
      word2parts = word2.split("/");
      // check if we really have two parts here
      if(word2parts.length != 2) {
        throw new GateRuntimeException("TreeTagger chunk tag does not contain 2 parts seperated by a slash: \""+line+"\"");
      }
      pos = word2parts[0];
      chunk = word2parts[1];
      Annotation token = (Annotation) tokens.get(i);   
      FeatureMap features = token.getFeatures();
      features.put("chunktag", chunk);
      part = chunk.substring(0,1);
      if(!part.equals("O") && chunk.length() > 2) {
        type = chunk.substring(2);
      } else { 
        type = "";
      }
      features.put("chunkpart",part);
      features.put("chunktype", type);
      if(getDoPosTagging().booleanValue()) {
        features.put("category",pos);
      }
      if(getIncludeLemma().booleanValue()) {
        features.put("lemma",words[2]);
      }
    }
  }
  /**
   * Set the flag for whether we we also want POS taggin tags
   */
  public void setDoPosTagging(Boolean newValue) {
    doPosTagging = (newValue == null) ? true : newValue.booleanValue();
  }

  /**
   * Get the flag for whether we should fail if an unmappable character is
   * found.
   */
  public Boolean getDoPosTagging() {
    return Boolean.valueOf(doPosTagging);
  }
  
  protected boolean doPosTagging;
 
  /**
   * Set the flag for whether we we also want POS taggin tags
   */
  public void setIncludeLemma(Boolean newValue) {
    includeLemma = (newValue == null) ? true : newValue.booleanValue();
  }

  /**
   * Get the flag for whether we should fail if an unmappable character is
   * found.
   */
  public Boolean getIncludeLemma() {
    return Boolean.valueOf(includeLemma);
  }
  
  protected boolean includeLemma;
 } // class TreeTaggerPOS
