package gate.treetagger2;

import java.util.*;

import gate.*;
import gate.creole.*;
import gate.util.*;

/** This class is a wrapper for the language-independent POS tagger from the
 * University of Stuttgart, Germany.
 * 
 * This class is for the POS-tagging function of TreeTagger.
 * 
 * This is a modified version of the plugin that comes with GATE version
 * 3.1 and earlier. This modified version includes several changes 
 * and enhancements over the original version:
 * <ul>
 * <li>The user does not have to specify the full path to the script
 * that runs TreeTagger. Instead, the plugin derives the location from
 * the location where the plugin is stored.
 * <li>The script that runs TreeTagger nearly never ever needs to be
 * changed since it uses several ways of determining how to invocate
 * the TreeTagger binaries and where to locate the TreeTagger parameter 
 * files (see below).
 * <li>The plugin does not read back the temporary TreeTagger input file
 * and store its content in temporary features of the Token annotations as 
 * the old version did, thus increasing CPU and memory efficiency a bit. 
 * The only check for consistency that is now done is
 * by comparing the number of Tokens with the number of lines read back
 * from the TreeTagger output which should be sufficient. 
 * <li>All plugin parameters are now runtime parameters so they can be
 * changed after creating the processing resource much more easily.
 * <li>The plugin now provides two processing resources: one for POS tagging
 * and one for chunking (provided a chunking model is available). This 
 * class handles the POS tagging.
 * <li>The plugin allows to specify an annotation type other than Token to
 * be used as input tot the TreeTagger program.
 * </ul>
 * 
 * The following parameters are available for the TreeTaggerPOS part of
 * speech tagger processing resource:
 * <ul>
 * <li> annotationSetName: the name of the annotation set to process</li>
 * <li> debugMode: true/false -- if true, will keep temporary files and show
 *   debugging messages</li>
 * <li> encoding: the encoding to use when creating the file that will be
 *   read in by the TreeTagger binary. This defaults to ISO-8859-1 and this
 *   is currently the only encoding supported by TreeTagger. </li>
 * <li> failOnUnmappableChar: what to do if the document contains a 
 *   character that cannot be converted to the specified encoding. If true,
 *   the processing will be interrupted, if false, the character will be 
 *   silently ignored.</li>
 * <li> tokenAnnotationType: the type name of the tokens to be used for 
 *   POS tagging. The default is "Token".</li>
 * <li> treeTaggerInvocationScriptParms: everything specified here is 
 *   passed on to the invocation script that actually locates and invokes
 *   the TreeTagger binary. The invocation script is located in the 
 *   plugin directory, subdirectory <code>cmd</code> and has the 
 *   name <code>run-treetagger.pl</code>.
 *   It can be invoked manually with the command 
 *   <code>perl cmd/run-treetagger.pl -h</code> to show all possible command
 *   line options and with the command 
 *   <code>perl cmd/run-treetagger.pl -man</code> to show the full documentation. 
 *   Note that some arguments do not apply to POS tagging and that 
 *   some arguments are passed automatically: <code>-tmpdir</code> and 
 *   <code>-chunk</code>. The only 
 *   argument that needs to be specified in most cases is <code>-lang</code>.
 *   </li>
 * </ul>
 *
 *  @author Johann Petrak, Austrian Research Institute for AI (OFAI)
 */

public class TreeTaggerPOS
    extends TreeTaggerBase
    implements ProcessingResource {

  static final long serialVersionUID = 20060730122746002L;
  
  public void execute() throws ExecutionException {
    super.execute();
  }
  
  /** Parse the lines of TreeTagger POS output and create features for
   *  the tokens.
   */
  protected void getFeatures4Tokens (ArrayList lines, ArrayList tokens) {
    String tag;
    String lemma;
    String line;
    for(int i=0;i<lines.size();i++) {
      line = (String) lines.get(i);
      String[] words = line.split("\\s+");
      if(words.length < 3) {
        throw new GateRuntimeException("A line returned by the TreeTagger contains less than 3 tokens: \""+line+"\"");
      }
      tag = words[1];
      lemma = words[2];
      Annotation token = (Annotation) tokens.get(i);   
      FeatureMap features = token.getFeatures();
      features.put("lemma", lemma);
      features.put("category", tag);
    }
  }
  
} // class TreeTaggerPOS
